!+
! Module: gtialign_mod
!
! Authors:  Dean Hinshaw - Leicester University
!           Ian Stewart (SSC/LUX)
!          
! Description: Align asynchronos GTI files to frame readout boundaries.
! 
! Comments: 
! 
! $Id: gtialign_mod.f90,v 1.2 2004/09/13 08:12:49 cgabriel Exp $
!- 
module gtialign_mod

  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

  public gtialign

  character(20) ::&
    style=''
  character(256) ::&
    inGtiSetTabName='',&
    eventListSetName='',&
    outGtiSetName='',&
    outGtiSetTabName='',&
    tsSetTabName=''

  character(256) ::&
    inGtiSetName='',&
    tsSetName=''
  character(20) ::&
    inGtiTabName='',&
    outGtiTabName='',&
    tsTabName=''

contains 
  
  !---------------------------------------------------------------------
  subroutine gtialign
  !---------------------------------------------------------------------
    
    implicit none 

    type(DataSetT)   :: eventListSet, tsSet, inGtiSet, outGtiSet
    type(BlockT)     :: inGtiBlock, eBlock
    type(TableT)     :: outGtiTable, inGtiTable, exposeTable, eventTable,&
                        tsTable
    integer          :: iBlock
    character(255)   :: blockName
    character(8)     :: newName
    logical          :: outside
    real(double), dimension(:), pointer :: inGtiStop
    real(double)     :: gtiZeroTime, gtiOffset
    real(single)     :: timePixR
    type(ColumnT)    :: inGtiStopCol

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    select case(style)
    case('pipeline')
      ! Initialize the output file

      eventListSet = dataSet(eventListSetName,Read)
      outGtiSet = dataSet(outGtiSetName,Create)

      call addAttributes(attributable(outGtiSet),attributable(eventListSet))

      ! Check the event list to see if stored frame times are at the beginning
      ! or end of the frame.  If so then a small offset needs to be added to the
      ! GTIs in order to ensure that the proper events are included/excluded.

      eventTable = table(eventListSet, "EVENTS")
      timePixR = 0.5
      gtiOffset = 0
      if( hasAttribute(eventTable, "TIMEPIXR") ) &
           timePixR = real32Attribute(eventTable, "TIMEPIXR")
      call release(eventTable)

      call splitSetTabName(inGtiSetTabName, inGtiSetName, inGtiTabName)
      inGtiSet = dataSet(inGtiSetName,Read)
      inGtiBlock = block(inGtiSet, inGtiTabName, Read)

      if( hasAttribute(inGtiBlock, "TIMEZERO") )then
         gtiZeroTime = real64Attribute(inGtiBlock, "TIMEZERO")
      else
         gtiZeroTime = 0
      end if

      inGtiTable = table(inGtiSet, inGtiTabName)
      if( timePixR<=0.0 .or. timePixR>=1.0 )then
         inGtiStopCol = column(inGtiTable, "STOP", READ)
         inGtiStop => real64Data(inGtiStopCol)

         gtiOffset = ( gtiZeroTime + inGtiStop(numberOfRows(inGtiTable)-1) )&
           * epsilon(gtiOffset)

         call release(inGtiStopCol)

         if( timePixR<=0.0 ) gtiOffset = -gtiOffset
      end if
      call release(inGtiTable)

      ! Loop thru the event file, looking for exposure extensions.  Create
      ! a gti extension for each found exposure extension.

      do iBlock=0, numberOfBlocks(eventListSet)-1
         eBlock = block(eventListSet, iBlock, READ)
         blockName = name(eBlock)
         call release(eBlock)

         if( blockName(1:6) == "EXPOSU" )then
            exposeTable = table(eventListSet, blockName)

            newName = "STDGTI"//blockName(7:8)
            call copyBlock(outGtiSet, inGtiBlock, newName)
            outGtiTable = table(outGtiSet, newName)

            call processGtis(exposeTable, outGtiTable, gtiZeroTime&
              , gtiOffset, outside)

            call release(outGtiTable)
            call release(exposeTable)
         end if
      end do

      call release(eventListSet)
      call release(outGtiSet)
      call release(inGtiBlock)
      call release(inGtiSet)

    case('generic')
      call splitSetTabName( inGtiSetTabName,  inGtiSetName,  inGtiTabName)
      call splitSetTabName(outGtiSetTabName, outGtiSetName, outGtiTabName)
      call splitSetTabName(    tsSetTabName,     tsSetName,     tsTabName)

      outGtiSet = dataSet(outGtiSetName, CREATE)
      tsSet = dataSet(tsSetName, READ)
      tsTable = table(tsSet, tsTabName)

      call addAttributes(attributable(outGtiSet),attributable(tsSet))

      inGtiSet = dataSet(inGtiSetName, READ)
      inGtiBlock = block(inGtiSet, inGtiTabName)

      if( hasAttribute(inGtiBlock, "TIMEZERO") )then
         gtiZeroTime = real64Attribute(inGtiBlock, "TIMEZERO")
      else
         gtiZeroTime = 0
      end if

      gtiOffset = 0.0d0

      call copyBlock(outGtiSet, inGtiBlock, outGtiTabName)
      outGtiTable = table(outGtiSet, outGtiTabName)

      call processGtis(tsTable, outGtiTable, gtiZeroTime&
        , gtiOffset, outside)

      call release(outGtiTable)
      call release(inGtiBlock)

      call release(inGtiSet)
      call release(tsTable)
      call release(tsSet)
      call release(outGtiSet)

    case default
call error('badStyle',"The value of the --style parameter was not recognized.")
    end select

    if( outside )then
       write(errstr, *) "Frame times do not span all GTIs.  GTI times &
         &outside of frame times are not modified."
       call warning("OutsideFrames")
    end if

  end subroutine gtialign
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    style = stringParameter('style')
    select case(style)
    case('pipeline')
      inGtiSetTabName  = stringParameter("gtitable")
      eventListSetName = stringParameter("eventset")
      outGtiSetName    = stringParameter("outset")
    case('generic')
      inGtiSetTabName  = stringParameter("ingtitable")
      outGtiSetTabName = stringParameter("outgtitable")
      tsSetTabName     = stringParameter("tstable")
    case default
call error('badStyle',"The value of the --style parameter was not recognized.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine splitSetTabName(setTabName, setName, tabName, noColonFound)
  !---------------------------------------------------------------------
!**** should access that in ssclib. However for the time being let's copy it to here.
  ! The sas param interface allows one to specify a dataset and table name
  ! in a single parameter, by using parameter type 'table'. The format of
  ! such a command-line argument is as in the followin example:
  !
  !	my_tab_parameter = fred.ds:BIGTAB
  !
  ! In practice it is easy inadvertently to leave off the colon + table
  ! name. In this case the dal does not offer any very helpful message.
  ! The aim of the present subroutine is (i) to allow people to leave off
  ! the tab name if they choose, in which case they get pointed to the
  ! first table in the dataset, and (ii) to issue a detailed and informatory
  ! warning in the case that it was necessary to specify the table name but
  ! the person forgot to do it.

    implicit none

    character(*),  intent(in)            :: setTabName
    character(*),  intent(out)           :: setName, tabName
    logical(bool), intent(out), optional :: noColonFound

    integer(int16) :: posOfColon
    type(DataSetT) :: set
    logical :: fileExists

    !-------------------------------------------------------------------
    ! Start

    posOfColon = index(setTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.

      inquire(file=setTabName, exist=fileExists)

      if (.not. fileExists) then
call error('missingBlockName', "No colon+block name was found after &
&the set name. I can't tell what block of the output file you meant to &
&access.")
      end if

      call warning('missingBlockName', "No colon+block name was &
        &found after the set name. I'm going to assume that you meant &
        &to access the first block in the set. If that's what you wanted, &
        &then there is no cause for alarm.")

      if (present(noColonFound)) noColonFound = .true.

      setName = setTabName
      set = dataSet(setName, READ)
      tabName = name(block(set, 0))
      call release(set)

    else
      if (present(noColonFound)) noColonFound = .false.

      setName = setTabName(:posOfColon - 1)
      tabName = setTabName(posOfColon + 1:)
    end if

  end subroutine splitSetTabName

  !---------------------------------------------------------------------
  subroutine processGtis(tsTab, outGtiTab, gtiZeroTime, gtiOffset, outside)
  !---------------------------------------------------------------------
  ! This contains the central algorithm of the original gtialign. Apart
  ! from extracting this algorithm into the present subroutine, I have
  ! made no changes to it apart from a few changes to variable names in
  ! the interests of greater clarity. IMS 20040907.

    implicit none 

    type(TableT), intent(in)  :: tsTab, outGtiTab
    real(double), intent(in)  :: gtiZeroTime, gtiOffset
    logical,      intent(out) :: outside

    integer :: iFrame, iGti, nFrames, iGood, nGoodRows, iBegin, iToRow, nGtiRows
    real(kind=double), dimension(:), pointer :: tsTimeData, gtiStart, gtiStop
    real(kind=single), dimension(:), pointer :: tsTimedelData
    real(kind=double) :: tsTimeZero
    real(kind=single) :: timePixR
    type(ColumnT) :: tsTimeCol, tsTimedelCol, gtiStartCol, gtiStopCol
    integer, allocatable, dimension(:) :: goodRows

    !-------------------------------------------------------------------
    ! Start:

    outside = .false. ! default
    nGtiRows = numberOfRows(outGtiTab)
    allocate(goodRows(0:nGtiRows) )

    tsTimeCol = column(tsTab, "TIME", READ)
    tsTimeData => real64Data(tsTimeCol)
    if( hasColumn(tsTab, "TIMEDEL") )then
      tsTimedelCol = column(tsTab, "TIMEDEL", READ)
      tsTimedelData => real32Data(tsTimedelCol)
    else
      allocate(tsTimedelData(0:numberOfRows(tsTab)))
      tsTimedelData = real64Attribute(tsTab, "TIMEDEL")
    end if
    if( hasAttribute(tsTab, "TIMEZERO") )then
      tsTimeZero = real64Attribute(tsTab, "TIMEZERO")
    else
      tsTimeZero = 0
    endif

    gtiStartCol = column(outGtiTab, "START", MODIFY)
    gtiStopCol  = column(outGtiTab, "STOP",  MODIFY)
    gtiStart => real64Data(gtiStartCol)
    gtiStop  => real64Data(gtiStopCol)

    timePixR = 0.5
    if( hasAttribute(tsTab, "TIMEPIXR") ) &
      timePixR = real32Attribute(tsTab, "TIMEPIXR")

    if( hasAttribute(tsTab,"CCDID") )&
      call setAttribute(outGtiTab, attribute(tsTab,"CCDID"))
    if( hasAttribute(tsTab,"CCDNODE") )&
      call setAttribute(outGtiTab, attribute(tsTab,"CCDNODE"))
    if( hasAttribute(tsTab,"FRMTIME") )&
      call setAttribute(outGtiTab, attribute(tsTab,"FRMTIME"))

    iFrame = 0
    nFrames = numberOfRows(tsTab)

    ! For each GTI, `align' the start and stop times with the frame boundaries:

    gtiLoop: do iGti=0, nGtiRows-1
      do while( tsTimeZero+tsTimeData(iFrame)-tsTimedelData(iFrame)*timePixR&
      < gtiZeroTime+gtiStart(iGti) )
        if (iFrame == nFrames - 1) then

          if (gtiZeroTime + gtiStart(iGti)&
          < tsTimeZero+tsTimeData(iFrame)+tsTimedelData(iFrame)*(1-timePixR))&
            gtiStart(iGti) = tsTimeZero + tsTimeData(iFrame)&
              + tsTimedelData(iFrame)*(1-timePixR) - gtiZeroTime + gtiOffset
    exit gtiLoop
        end if
        iFrame = iFrame + 1
      end do
      if (iFrame == 0) then
        if( gtiZeroTime+gtiStop(iGti)&
        < tsTimeZero+tsTimeData(iFrame)-tsTimedelData(iFrame)*timePixR) then
          outside = .true.
    cycle gtiLoop
        end if
      else
        gtiStart(iGti) = tsTimeZero + tsTimeData(iFrame)&
          - tsTimedelData(iFrame) * timePixR - gtiZeroTime + gtiOffset
      end if

      do while (tsTimeZero + tsTimeData(iFrame) + tsTimedelData(iFrame)&
      * (1-timePixR) < gtiZeroTime + gtiStop(iGti))
        if( iFrame == nFrames - 1 )then
          outside = .true.
    exit gtiLoop
        end if
        iFrame = iFrame + 1
      end do
      if (iFrame == 0) then
        gtiStop(iGti) = tsTimeZero + tsTimeData(iFrame)&
          - tsTimedelData(iFrame) * timePixR - gtiZeroTime + gtiOffset
      else
        gtiStop(iGti) = tsTimeZero + tsTimeData(iFrame-1)&
          + tsTimedelData(iFrame-1) * (1-timePixR) - gtiZeroTime + gtiOffset
      end if
    end do gtiLoop

    ! Find and delete all gti rows which do no span a frame boundary.
    ! These show up as rows where START > STOP

    nGoodRows = 0

    do iGti=0, nGtiRows-1
      if( gtiStart(iGti) < gtiStop(iGti) )then
        goodRows(nGoodRows) = iGti
        nGoodRows = nGoodRows+1
      end if
    end do

    call release(gtiStartCol)
    call release(gtiStopCol)

    iBegin = 0
    iToRow = 0
    do iGood=0, nGoodRows-1
      if( goodRows(iGood) == nGtiRows-1&
      .or. goodRows(iGood)+1 /= goodRows(iGood+1)) then
        call copyRows(outGtiTab, goodRows(iBegin), iToRow, iGood-iBegin+1)
        iToRow = iToRow+iGood-iBegin+1
        iBegin = iGood+1
      end if
    end do

    call deleteRows(outGtiTab, iToRow, nGtiRows-iToRow)

  end subroutine processGtis

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module gtialign_mod
